using NBitcoin;
using Nostr.Client.Identifiers;
using Nostr.Client.Json;
using Nostr.Client.Messages;
using Nostr.Client.Messages.Metadata;
using Nostr.Client.Utils;
using ProtoBuf;
using StackExchange.Redis;

namespace NostrServices.Services;

public class RedisStore
{
    private static readonly TimeSpan DefaultExpire = TimeSpan.FromDays(90);
    private readonly IDatabase _database;

    public RedisStore(IDatabase database)
    {
        _database = database;
    }

    public async Task<bool> StoreEvent(CompactEvent ev, TimeSpan? expiry = null)
    {
        return await _database.SetAsync(EventKey(ev.ToIdentifier()), ev, expiry ?? DefaultExpire);
    }

    public async Task<CompactEvent?> GetEvent(NostrIdentifier id)
    {
        var ek = EventKey(id);
        var ev = await _database.GetAsync<CompactEvent>(ek);
        if (ev != default)
        {
            await _database.KeyExpireAsync(ek, DefaultExpire);
        }

        return ev;
    }

    public async Task StoreProfile(CompactProfile meta, TimeSpan? expiry = null)
    {
        var oldProfile = await GetProfile(meta.PubKey.ToHex());
        if ((oldProfile?.Created ?? new DateTime()) < meta.Created)
        {
            await _database.SetAsync(ProfileKey(meta.PubKey.ToHex()), meta, expiry ?? DefaultExpire);
        }
    }

    public async Task<CompactProfile?> GetProfile(string id)
    {
        var pk = ProfileKey(id);
        var profile = await _database.GetAsync<CompactProfile>(pk);
        if (profile != default)
        {
            await _database.KeyExpireAsync(pk, DefaultExpire);
        }

        return profile;
    }

    private string EventKey(NostrIdentifier id)
    {
        if (id is NostrAddressIdentifier naddr)
        {
            return $"event:{naddr.Kind}:{naddr.Pubkey}:{naddr.Identifier}";
        }

        return $"event:{id.Special}";
    }

    private string ProfileKey(string id) => $"profile:{id}";
}

[ProtoContract]
public class CompactEventTag
{
    [ProtoMember(1)]
    public string Key { get; init; } = null!;

    [ProtoMember(2)]
    public List<string> Values { get; init; } = null!;
}

[ProtoContract]
public class CompactEvent
{
    [ProtoMember(1)]
    public byte[] Id { get; init; } = null!;

    [ProtoMember(2)]
    public byte[] PubKey { get; init; } = null!;

    [ProtoMember(3)]
    public long Created { get; init; }

    [ProtoMember(4)]
    public string Content { get; init; } = null!;

    [ProtoMember(5)]
    public List<CompactEventTag> Tags { get; init; } = new();

    [ProtoMember(6)]
    public byte[] Sig { get; init; } = null!;

    [ProtoMember(7)]
    public long Kind { get; init; }

    public static CompactEvent FromNostrEvent(NostrEvent ev)
    {
        return new CompactEvent
        {
            Id = Convert.FromHexString(ev.Id!),
            PubKey = Convert.FromHexString(ev.Pubkey!),
            Kind = (long)ev.Kind,
            Created = ev.CreatedAt!.Value.ToUnixTimestamp(),
            Content = ev.Content!,
            Tags = ev.Tags!.Select(a => new CompactEventTag
            {
                Key = a.TagIdentifier!,
                Values = a.AdditionalData.Cast<string>().ToList()
            }).ToList(),
            Sig = Convert.FromHexString(ev.Sig!)
        };
    }

    public NostrEvent ToNostrEvent()
    {
        return new NostrEvent()
        {
            Id = Id.ToHex(),
            Pubkey = PubKey.ToHex(),
            CreatedAt = DateTimeOffset.FromUnixTimeSeconds(Created).UtcDateTime,
            Content = Content,
            Tags = new(Tags.Select(a => new NostrEventTag(a.Key, a.Values.ToArray()))),
            Sig = Sig.ToHex()
        };
    }
}

[ProtoContract]
public class CompactProfile
{
    [ProtoMember(1)]
    public byte[] PubKey { get; init; } = null!;

    [ProtoMember(2)]
    public string? Name { get; init; }

    [ProtoMember(3)]
    public string? About { get; init; }

    [ProtoMember(4)]
    public string? Picture { get; init; }

    [ProtoMember(5)]
    public string? Nip05 { get; init; }

    [ProtoMember(6)]
    public string? Lud16 { get; init; }

    [ProtoMember(7)]
    public string? Banner { get; init; }

    [ProtoMember(8)]
    public DateTime Created { get; init; }

    public static CompactProfile? FromNostrEvent(NostrEvent ev)
    {
        var meta = NostrJson.Deserialize<NostrMetadata>(ev.Content);
        if (meta != default)
        {
            return new()
            {
                PubKey = Convert.FromHexString(ev.Pubkey!),
                Name = meta.Name,
                About = meta.About,
                Picture = meta.Picture,
                Nip05 = meta.Nip05,
                Lud16 = meta.Lud16,
                Banner = meta.Banner,
                Created = ev.CreatedAt!.Value
            };
        }

        return default;
    }
}
